#!/usr/bin/env python

"""Main script."""

import json
import msgpack
import pandas as pd
from pathlib import Path
from datetime import datetime

from flask import Flask, send_file, make_response
from flask_restful import Resource, Api
from loguru import logger

try:
    from vicon_dssdk import ViconDataStream
except ImportError:
    logger.error("Make sure vicon DataStreamSDK is installed: Follow the instructions in https://www.vicon.com/software/datastream-sdk/\n")
    #raise

def get_client(connection=None):
    if connection is None:
        connection = 'localhost:801'
    client = ViconDataStream.Client()

    logger.info('Connecting...')
    while not client.IsConnected():
        client.Connect(connection)
    logger.info('Connected to vicon data stream')
    client.EnableSegmentData()
    client.EnableMarkerData()
##    client.SetAxisMapping(ViconDataStream.Client.AxisMapping.EForward,
##                          ViconDataStream.Client.AxisMapping.EUp,
##                          ViconDataStream.Client.AxisMapping.ELeft)
    logger.info(client.GetAxisMapping())
    return client


def process_return_value(ret_val, use_json=False):
    if use_json:
        return ret_val
    else:
        if isinstance(ret_val, tuple):
            response = make_response(msgpack.packb(ret_val[0]), ret_val[1])
        else:
            response = make_response(msgpack.packb(ret_val))
        response.headers['content-type'] = 'application/msgpack'
        return response


def _init_api(connection=None, host="127.0.0.1", port="5000", use_json=False):
    try:
        client = get_client(connection)
    except Exception as e:
        logger.exception("Failed to connect to client")
        client = None
    app = Flask("vicon-ds")
    api = Api(app)

    class ViconMarkerStream(Resource):
        def get(self, data_type, subject_name):
            ret_val = self._get(data_type, subject_name)
            return process_return_value(ret_val, use_json)

        def _get(self, data_type, subject_name):
            if client is not None and client.IsConnected() and client.GetFrame():
                return get_data(client, data_type, subject_name)
            return "restart:  client didn't connect", 404

    api.add_resource(ViconMarkerStream, '/<string:data_type>/<string:subject_name>')
    try:
        app.run(host=host, port=int(port))
    finally:
        pass


LINES = []
IDX = 0
PLAY_MODE = False
PLAY_INDEX = None
PLAY_TS = None

def _init_api_static(connection=None, host="127.0.0.1", port="5000", input_file=None, use_json=False):
    app = Flask("vicon-ds")
    api = Api(app)

    if input_file is None or len(input_file) == 0:
        logger.error("`input_file` cannot be empty")
        return

    _lines = []
    with open(input_file) as f:
        for l in f.readlines():
            if len(l.strip()) == 0:
                continue
            _lines.append(l.rstrip().split(",", maxsplit=1))

    global LINES
    LINES = pd.DataFrame(_lines)
    LINES[1] = LINES[1].apply(lambda x: x if len(x) > 0 else None)
    LINES = LINES.dropna()
    LINES[0] = pd.to_numeric(LINES[0])

    # TODO: better validation?
    class ViconMarkerStreamProcess(Resource):
        def _set_index(self, idx):
            global IDX
            if idx >= LINES.index.max():
                IDX = 0
            elif idx < 0:
                IDX = int(LINES.index.max())
            else:
                IDX = idx

        def get(self, process=None, param=None):
            ret_val = self._get(process, param)
            return process_return_value(ret_val, use_json=True)

        def _get(self, process=None, param=None):
            global IDX, PLAY_MODE, PLAY_INDEX, PLAY_TS
            if process is None or process == "index":
                return send_file(Path(__file__).parent / "static" / "index.html")
            elif process == "n":
                self._set_index(IDX + 1)
                return IDX
            elif process == "p":
                self._set_index(IDX - 1)
                return IDX
            elif process == "s":
                if param is None:
                    return "param cannot be empty. Use: /offline/s/<frame-number>", 404
                try:
                    self._set_index(int(param))
                    return IDX
                except:
                    return "param should be a number. Use: /offline/s/<frame-number>", 404
            elif process == "t":
                PLAY_MODE = not PLAY_MODE
                if PLAY_MODE:
                    PLAY_INDEX = LINES.iloc[IDX, 0]
                    PLAY_TS = datetime.now().timestamp()
                else:
                    self._set_index(int(LINES[LINES.iloc[:, 0] == PLAY_INDEX].index[0]))
                return PLAY_MODE
            return "Process not recognized. Available processes: offline/n  = Next, offline/p = Previous, offline/s/<frame-number> = jump to frame-number, offline/t = toggle play mode", 404

    class ViconMarkerStream(Resource):
        def get(self, data_type, subject_name):
            ret_val = self._get(data_type, subject_name)
            return process_return_value(ret_val, use_json)

        def _get(self, data_type, subject_name):
            global PLAY_INDEX, PLAY_TS
            if subject_name == 'test':
                if PLAY_MODE:
                    index = LINES.iloc[:, 0]
                    diff = (index - PLAY_INDEX)
                    min_val = index[diff > 0].min()
                    ts = datetime.now().timestamp()
                    if ts - PLAY_TS >= 0:
                        PLAY_INDEX = min_val
                        PLAY_TS = ts
                    if min_val == index.max():
                        PLAY_INDEX = index.min()
                        PLAY_TS = datetime.now().timestamp()
                    return json.loads(LINES[index == PLAY_INDEX].iloc[0, 1])
                else:
                    return json.loads(LINES.iloc[IDX, 1])
            return "Only works for subject `test`", 404

    api.add_resource(ViconMarkerStream, '/<string:data_type>/<string:subject_name>')
    api.add_resource(ViconMarkerStreamProcess, '/offline/<string:process>', '/offline/<string:process>/<string:param>', '/offline')
    print(f"Go to `http://{host}:{port}/offline/index` to access the web UI")
    app.run(host=host, port=int(port))
            

def get_data(client, data_type, subject_name):
    data = {}
    # logger.info(*[n for n in client.__dir__() if "G" in n], sep="\n")
    # slogger.info(client.GetSegmentNames(subject_name))
    if data_type == "marker":
        marker_segment_data = {}
        marker_data = {}
        for marker, segment in client.GetMarkerNames(subject_name):
            try:
                marker_segment_data[segment].append(marker)
            except KeyError:
                marker_segment_data[segment] = [marker]
            marker_data[marker] = client.GetMarkerGlobalTranslation(subject_name, marker)[0]
            # logger.info(client.GetMarkerGlobalTranslation(subject_name, marker))
        data['data'] = marker_data
        data['hierachy'] = marker_segment_data
        
    elif data_type == "segment":
        segment_data = {}
        for segment in client.GetSegmentNames(subject_name):
            translation, status = client.GetSegmentGlobalTranslation(subject_name, segment)
            rotation, status = client.GetSegmentGlobalRotationQuaternion(subject_name, segment)
            segment_data[segment] = translation + rotation
        data['data'] = segment_data
            
    return data


def main(connection=None):
    client = get_client(connection)
    try:
        while client.IsConnected():
            if client.GetFrame():
                logger.info(get_data(client, 'test'))

    except ViconDataStream.DataStreamException as e:
        logger.error( f'Error: {e}' )


if __name__ == '__main__':  # pragma: no cover
    main()



"""CLI."""

import click

from ._main import main, _init_api, _init_api_static

@click.group()
def _main():
    pass


@click.command()
@click.option("-c", "--connection", default='localhost:801')
def test(connection):
    """Test if connection is working"""
    main(connection)


@click.command()
@click.option("-c", "--connection", default='localhost:801')
@click.option("-h", "--host", default='127.0.0.1')
@click.option("-p", "--port", default='5000')
@click.option("-j", "--use-json", default=False, is_flag=True)
def server(connection, host, port, use_json):
    """Connects to the vicon and streams the data out through host:port"""
    _init_api(connection=connection, host=host, port=port, use_json=use_json)


@click.command()
@click.option("-c", "--connection", default='localhost:801')
@click.option("-h", "--host", default='127.0.0.1')
@click.option("-p", "--port", default='5000')
@click.option("-f", "--file", default=None)
@click.option("-j", "--use-json", default=False, is_flag=True)
def stream(connection, host, port, file, use_json):
    """
    Instead of connecting to vicon, streams data from a csv file. The format of the file is
    the same as the one recorded by https://github.com/ahmed-shariff/vicon-nexus-unity-stream.
    """
    _init_api_static(connection=connection, host=host, port=port, input_file=file, use_json=use_json)


_main.add_command(test)
_main.add_command(server)
_main.add_command(stream)
    

if __name__ == '__main__':  # pragma: no cover
    _main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vicon_nexus_unity_stream_py', 'vicon_nexus_unity_stream_py.tests']

package_data = \
{'': ['*'], 'vicon_nexus_unity_stream_py': ['static/*']}

install_requires = \
['Flask-RESTful>=0.3.9,<0.4.0',
 'click>=8.1.3,<9.0.0',
 'flask_restful>=0.3.8,<0.4.0',
 'loguru>=0.5.0,<0.6.0',
 'msgpack>=1.0.4,<2.0.0',
 'pandas>=1.5.1,<2.0.0']

entry_points = \
{'console_scripts': ['vicon-nexus-stream = '
                     'vicon_nexus_unity_stream_py.cli:_main']}

setup_kwargs = {
    'name': 'vicon-nexus-unity-stream-py',
    'version': '0.2',
    'description': 'Python script to stream data from vicon nexus to unity',
    'long_description': '# Overview\n\nPython script to stream data from vicon nexus to unity\n\n\n# Setup\n\n## Requirements\n\n* Python 3.8+\n\n## Installation\n\nInstall it directly into an activated virtual environment:\n\n```text\n$ pip install vicon_nexus_unity_stream_py\n```\n\nor add it to your [Poetry](https://poetry.eustace.io/) project:\n\n```text\n$ poetry add vicon_nexus_unity_stream_py\n```\n\n# Usage\n\nAfter installation, the package can be used as a cli tool:\n\n```text\n$ vicon-nexus-stream --help\n\nUsage: vicon-nexus-stream [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  server  Connects to the vicon and streams the data...\n  stream  Instead of connecting to vicon, streams data...\n  test    Test if connection is working\n```\n\n\n# Credits\nThis project was generated with [cookiecutter](https://github.com/audreyr/cookiecutter) using [jacebrowning/template-python](https://github.com/jacebrowning/template-python).\n',
    'author': 'Ahmed Shariff',
    'author_email': 'shariff.mfa@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/vicon_nexus_unity_stream_py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
